#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include <3069S.H>
#include <mon_func.h>

#define CLOCKS_PER_SEC (25*1000*1000/8)
#define MAXN 100
int x[MAXN];
int startn = 100;
int n;

/* FUNCTIONS TO BE TIMED */

int intcmp(int *i, int *j)
{
    return *i - *j;
}

#define swapmac(i, j) { t = x[i]; x[i] = x[j]; x[j] = t; }

void swapfunc(int i, int j)
{
    int t = x[i];
    x[i] = x[j];
    x[j] = t;
}

#define maxmac(a, b) ((a) > (b) ? (a) : (b))

int maxfunc(int a, int b)
{
    return a > b ? a : b;
}

/* WORKHORSE */

#define T(s) printf("%s (n=%d)\r\n", s, n);
#define TRIALS 5
#define M(op)                                           \
	printf(" %-22s", #op);                          \
	k = 0;                                          \
	timesum = 0;                                    \
	for (ex = 0; ex < TRIALS; ex++) {               \
		start = clock();                        \
		for (i = 1; i <= n; i++) {              \
			fi = (float) i;                 \
			for (j = 1; j <= n; j++) {      \
				op;                     \
			}                               \
		}                                       \
		t = clock()-start;                      \
		timesum += t;                           \
	}                                               \
	nans = 1e9 * timesum / ((double)                \
		n*n * TRIALS * CLOCKS_PER_SEC);         \
	printf("%8.0f\r\n", nans);

static int cnt=0;

void ovf_func()
{
    cnt++;
    ITU.TISRC.BIT.OVF0=0;
}

long long int clock()
{
    return (cnt<<16)+ITU0.TCNT;
}

void nmi_handle(void)
{
    WDT.WRITE.TCSR=0xa5f8;
    while(1);
}

int main()
{
    volatile int i, j, k;
    volatile float fi, fj, fk;
    volatile long long int t, ex, timesum, start, globalstart;
    volatile double nans;

    int_regist(7,nmi_handle);

    ITU0.TCR.BIT.TPSC=3;
    ITU0.TCR.BIT.CCLR=1;
    int_regist(26,ovf_func);
    ITU.TSTR.BIT.STR0=1;
    ITU.TISRC.BIT.OVIE0=1;

    globalstart = clock();

    for (i = 0; i < MAXN; i++)
        x[i] = clock();
    n = startn;
    printf("C Time Cost Model, n=%d\r\n", n);

    T("Integer Arithmetic");
    M({});
    M(k++);
    M(k = i + j);
    M(k = i - j);
    M(k = i * j);
    M(k = i / j);
    M(k = i % j);
    M(k = i & j);
    M(k = i | j);
    T("Floating Point Arithmetic");
    M(fj=j;);
    M(fj=j; fk = fi + fj;);
    M(fj=j; fk = fi - fj;);
    M(fj=j; fk = fi * fj;);
    M(fj=j; fk = fi / fj;);
    T("Array Operations");
    M(k = i + j);
    M(k = x[i] + j);
    M(k = i + x[j]);
    M(k = x[i] + x[j]);
    T("Comparisons");
    M(if (i < j) k++);
    M(if (x[i] < x[j]) k++);
    T("Array Comparisons and Swaps");
    M(k = (x[i]<x[k]) ? -1:1);
    M(k = intcmp(x+i, x+j));
    M(swapmac(i, j));
    M(swapfunc(i, j));
    T("Max Function, Macro and Inline");
    M(k = (i > j) ? i : j);
    M(k = maxmac(i, j));
    M(k = maxfunc(i, j));
    n = startn / 10;
    T("Math Functions");
    M(fk = j+fi;);
    M(fk = sqrt(j+fi));
    M(fk = sin(j+fi));
    M(fk = sinh(j+fi));
    M(fk = asin(j+fi));
    M(fk = cos(j+fi));
    M(fk = tan(j+fi));
    n = startn;
    T("Memory Allocation");
    M(free(malloc(16)););
    M(free(malloc(100)););
    M(free(malloc(2000)););

/* Possible additions: input, output, malloc */
    printf("  Secs: %4.2f\r\n",
           ((double) clock()-globalstart)
           / CLOCKS_PER_SEC);
    return 0;
}
